#include "nicontrol.h"
#include "frame.h"
#include <QDebug>

#define MAX_DEPTH 10000

#define MIN_NUM_CHUNKS(data_size, chunk_size)	((((data_size)-1) / (chunk_size) + 1))
#define MIN_CHUNKS_SIZE(data_size, chunk_size)	(MIN_NUM_CHUNKS(data_size, chunk_size) * (chunk_size))

// NOTE: Rever o uso deste define
#define TEXTURE_SIZE	1


NIControl * NIControl::instance = NULL;

NIControl::NIControl(QString oniFile)
    : m_rc(openni::STATUS_OK)
    , m_eViewState(DISPLAY_MODE_OVERLAY)
{
    zMin=0; zMax=6000;
    frame = NULL;
    m_pTexMap = NULL;
    setup(oniFile);
}

NIControl::~NIControl()
{
    stopCamera();

    delete[] m_pTexMap;
    delete[] m_pDepthMap;

    if (m_streams != NULL)
    {
        delete []m_streams;
    }


}

NIControl * NIControl::getInstance(QString oniFile)
{
    if( !instance )
    {
        instance = new NIControl(oniFile);
    }

    return instance;
}

void NIControl::initOpenNI(QString oniFile)
{

    if(!oniFile.isEmpty()) {
        inputMode = FromFile;
        this->oniFile = oniFile;
    } else {
        inputMode = FromDevice;
    }


    m_rc = openni::OpenNI::initialize();

    printf("After initialization:\n%s\n", openni::OpenNI::getExtendedError());


    if(inputMode == FromFile)
        m_rc = m_device.open(this->oniFile.toStdString().c_str());
    else
        m_rc = m_device.open(openni::ANY_DEVICE);

    if (m_rc != openni::STATUS_OK)
    {
        printf("SimpleViewer: Device open failed:\n%s\n", openni::OpenNI::getExtendedError());
        openni::OpenNI::shutdown();
        return;
    }

    m_rc = m_depthStream.create(m_device, openni::SENSOR_DEPTH);
    if (m_rc == openni::STATUS_OK)
    {
        m_rc = m_depthStream.start();
        if (m_rc != openni::STATUS_OK)
        {
            printf("SimpleViewer: Couldn't start depth stream:\n%s\n", openni::OpenNI::getExtendedError());
            m_depthStream.destroy();
        }
    }
    else
    {
        printf("SimpleViewer: Couldn't find depth stream:\n%s\n", openni::OpenNI::getExtendedError());
    }

    m_rc = m_colorStream.create(m_device, openni::SENSOR_COLOR);
    if (m_rc == openni::STATUS_OK)
    {
        m_rc = m_colorStream.start();
        if (m_rc != openni::STATUS_OK)
        {
            printf("SimpleViewer: Couldn't start color stream:\n%s\n", openni::OpenNI::getExtendedError());
            m_colorStream.destroy();
        }
    }
    else
    {
        printf("SimpleViewer: Couldn't find color stream:\n%s\n", openni::OpenNI::getExtendedError());
    }

    if (!m_depthStream.isValid() || !m_colorStream.isValid())
    {
        printf("SimpleViewer: No valid streams. Exiting\n");
        openni::OpenNI::shutdown();
        return;
    }

    if(m_device.isImageRegistrationModeSupported(openni::IMAGE_REGISTRATION_DEPTH_TO_COLOR))
    {
        m_device.setImageRegistrationMode(openni::IMAGE_REGISTRATION_DEPTH_TO_COLOR);
    }

    started = initCamera();

    return;
}

bool NIControl::initCamera()
{
    openni::VideoMode depthVideoMode;
    openni::VideoMode colorVideoMode;
    int width,height;

    if (m_depthStream.isValid() && m_colorStream.isValid())
    {
        depthVideoMode = m_depthStream.getVideoMode();
        colorVideoMode = m_colorStream.getVideoMode();

        int depthWidth = depthVideoMode.getResolutionX();
        int depthHeight = depthVideoMode.getResolutionY();
        int colorWidth = colorVideoMode.getResolutionX();
        int colorHeight = colorVideoMode.getResolutionY();

        if (depthWidth != colorWidth || depthHeight != colorHeight)
        {
            printf("Error - expect color and depth to be in same resolution: D: %dx%d, C: %dx%d\n",
                depthWidth, depthHeight,
                colorWidth, colorHeight);
            m_rc = openni::STATUS_ERROR;
        }
        width = depthWidth;
        height = depthHeight;
    }
    else if (m_depthStream.isValid())
    {
        depthVideoMode = m_depthStream.getVideoMode();
    }
    else if (m_colorStream.isValid())
    {
        colorVideoMode = m_colorStream.getVideoMode();
    }
    else
    {
        printf("Error - expects at least one of the streams to be valid...\n");
        m_rc = openni::STATUS_ERROR;
    }

    if (m_rc != openni::STATUS_OK)
    {
        openni::OpenNI::shutdown();
        return false;
    }

    m_streams = new openni::VideoStream*[2];
    m_streams[0] = &m_depthStream;
    m_streams[1] = &m_colorStream;

    // Texture map init
    m_nTexMapX = MIN_CHUNKS_SIZE(width, TEXTURE_SIZE);
    m_nTexMapY = MIN_CHUNKS_SIZE(height, TEXTURE_SIZE);
    m_pTexMap = new openni::RGB888Pixel[m_nTexMapX * m_nTexMapY];

    // Depth map init
    m_pDepthMap = new openni::DepthPixel[m_nTexMapX * m_nTexMapY];

    return true;

}

bool NIControl::stopCamera()
{
    m_depthStream.stop();
    m_colorStream.stop();
    m_depthStream.destroy();
    m_colorStream.destroy();
    m_device.close();
    openni::OpenNI::shutdown();
    return true;
}

void NIControl::calculateHistogram(float *pHistogram, int histogramSize, const openni::VideoFrameRef &frame)
{
    const openni::DepthPixel* pDepth = (const openni::DepthPixel*)frame.getData();
    // Calculate the accumulative histogram (the yellow display...)
    memset(pHistogram, 0, histogramSize*sizeof(float));
    int restOfRow = frame.getStrideInBytes() / sizeof(openni::DepthPixel) - frame.getWidth();
    int height = frame.getHeight();
    int width = frame.getWidth();

    unsigned int nNumberOfPoints = 0;
    for (int y = 0; y < height; ++y)
    {
        for (int x = 0; x < width; ++x, ++pDepth)
        {
            if (*pDepth != 0)
            {
                pHistogram[*pDepth]++;
                nNumberOfPoints++;
            }
        }
        pDepth += restOfRow;
    }
    for (int nIndex=1; nIndex<histogramSize; nIndex++)
    {
        pHistogram[nIndex] += pHistogram[nIndex-1];
    }
    if (nNumberOfPoints)
    {
        for (int nIndex=1; nIndex<histogramSize; nIndex++)
        {
            pHistogram[nIndex] = (256 * (1.0f - (pHistogram[nIndex] / nNumberOfPoints)));
        }
    }
}

Frame* NIControl::getFrame()
{

    int changedIndex;
    openni::Status rc = openni::OpenNI::waitForAnyStream(m_streams, 2, &changedIndex);
    if (rc != openni::STATUS_OK)
    {
        printf("Wait failed\n");
        return frame;
    }

    switch (changedIndex)
    {
    case 0:
        m_depthStream.readFrame(&m_depthFrame); break;
    case 1:
        m_colorStream.readFrame(&m_colorFrame); break;
    default:
        printf("Error in wait\n");
    }

    if (m_depthFrame.isValid())
    {
        calculateHistogram(m_pDepthHist, MAX_DEPTH, m_depthFrame);
    }

    memset(m_pTexMap, 0, m_nTexMapX*m_nTexMapY*sizeof(openni::RGB888Pixel));
    memset(m_pDepthMap, 0, m_nTexMapX*m_nTexMapY*sizeof(uint16_t));

    // check if we need to draw image frame to texture
    if ((m_eViewState == DISPLAY_MODE_OVERLAY ||
         m_eViewState == DISPLAY_MODE_IMAGE) && m_colorFrame.isValid())
    {
        const openni::RGB888Pixel* pImageRow = (const openni::RGB888Pixel*)m_colorFrame.getData();
        openni::RGB888Pixel* pTexRow = m_pTexMap + m_colorFrame.getCropOriginY() * m_nTexMapX;
        int rowSize = m_colorFrame.getStrideInBytes() / sizeof(openni::RGB888Pixel);

        for (int y = 0; y < m_colorFrame.getHeight(); ++y)
        {
            const openni::RGB888Pixel* pImage = pImageRow;
            openni::RGB888Pixel* pTex = pTexRow + m_colorFrame.getCropOriginX();

            for (int x = 0; x < m_colorFrame.getWidth(); ++x, ++pImage, ++pTex)
            {
                *pTex = *pImage;
            }

            pImageRow += rowSize;
            pTexRow += m_nTexMapX;
        }
    }

    if (m_depthFrame.isValid())
    {
        const openni::DepthPixel* pDepthRow = (const openni::DepthPixel*)m_depthFrame.getData();

        // check if we need to draw depth frame to texture
        if ((m_eViewState == DISPLAY_MODE_OVERLAY ||
             m_eViewState == DISPLAY_MODE_DEPTH))
        {
            openni::RGB888Pixel* pTexRow = m_pTexMap + m_depthFrame.getCropOriginY() * m_nTexMapX;
            int rowSize = m_depthFrame.getStrideInBytes() / sizeof(openni::DepthPixel);

            for (int y = 0; y < m_depthFrame.getHeight(); ++y)
            {
                const openni::DepthPixel* pDepth = pDepthRow;
                openni::RGB888Pixel* pTex = pTexRow + m_depthFrame.getCropOriginX();

                for (int x = 0; x < m_depthFrame.getWidth(); ++x, ++pDepth, ++pTex)
                {
                    if (*pDepth != 0)
                    {
                        int nHistValue = m_pDepthHist[*pDepth];
                        pTex->r = nHistValue;
                        pTex->g = nHistValue;
                        pTex->b = 0;
                    }
                }

                pDepthRow += rowSize;
                pTexRow += m_nTexMapX;
            }
        }
    }

    frame->pTexMap = m_pTexMap;
    frame->pRGBMap = m_pTexMap;
    frame->pDepthMap = m_pDepthMap;

    return frame;
}

void NIControl::setup(QString fileName)
{

    if(frame != NULL) {
        delete frame;
        frame = NULL;
    }

    if(m_pTexMap != NULL) {
        delete [] m_pTexMap;
        m_pTexMap = NULL;
    }

    stopCamera();
    initOpenNI(fileName);
    depthWidth = m_depthStream.getVideoMode().getResolutionX();
    depthHeight = m_depthStream.getVideoMode().getResolutionY();
    m_pTexMap = (openni::RGB888Pixel*)malloc(depthWidth * depthHeight * sizeof(openni::RGB888Pixel));

    frame = new Frame(depthWidth, depthHeight);

}

