#include "nicontrol.h"
#include "frame.h"
#include <QDebug>

#define SAMPLE_XML_PATH "/Users/lvelho/Documents/vgip/config/SamplesConfig.xml"
#define MAX_DEPTH 10000

NIControl * NIControl::instance = NULL;

NIControl::NIControl(QString oniFile) :
    pDepthHist(new float[MAX_DEPTH])
{
    zMin=0; zMax=6000;
    frame = NULL;
    pTexMap = NULL;
    setup(oniFile);
}

NIControl::~NIControl()
{
}

NIControl * NIControl::getInstance(QString oniFile)
{
    if( !instance )
    {
        instance = new NIControl(oniFile);
    }

    return instance;
}

void NIControl::initOpenNI(QString oniFile)
{
    XnStatus rc;
    xn::EnumerationErrors errors;

    if(!oniFile.isEmpty()) {
        inputMode = FromFile;
        this->oniFile = oniFile;
    } else {
        inputMode = FromDevice;
    }

    if(inputMode == FromFile) {
        rc = niContext.Init();
        if(rc != XN_STATUS_OK) {
            fprintf(stderr, "Failed to initialize OpenNI: %s\n", xnGetStatusString(rc));
            exit(1);
        }

        // Open file recording
        rc = niContext.OpenFileRecording(oniFile.toLatin1().data());
        if(rc != XN_STATUS_OK) {
            fprintf(stderr,"Failed to open the capture file %s.\n", oniFile.toLatin1().data());
            started = false;
            return;
        }

    } else {

        rc = niContext.InitFromXmlFile(SAMPLE_XML_PATH, &errors);
    }

    if(rc == XN_STATUS_NO_NODE_PRESENT)
    {
        XnChar strError[1024];
        errors.ToString(strError, 1024);
        qDebug() << "[NICONTROL] " << strError;
        started = false;
        return;
    }
    else if(rc != XN_STATUS_OK)
    {
        qDebug() << "[NICONTROL] Open failed: " << xnGetStatusString(rc) << oniFile;
        started = false;
        return;
    }

    rc = niContext.FindExistingNode(XN_NODE_TYPE_DEPTH, depth);
    rc = niContext.FindExistingNode(XN_NODE_TYPE_IMAGE, image);

    depth.GetMetaData(depthMD);
    image.GetMetaData(imageMD);

    if (depth.IsCapabilitySupported(XN_CAPABILITY_ALTERNATIVE_VIEW_POINT)) {
        depth.GetAlternativeViewPointCap().SetViewPoint(image);
        qDebug() << "Using internal calibration.";
    }


    if(imageMD.FullXRes() != depthMD.FullXRes() || imageMD.FullYRes() != depthMD.FullYRes())
    {
        qDebug() << "The device depth and image resolution must be equal!";
    }

    if(imageMD.PixelFormat() != XN_PIXEL_FORMAT_RGB24)
    {
        qDebug() << "The device image format must be RGB24";
    }

    started = true;
    return;
}

Frame* NIControl::getFrame()
{

    XnStatus rc = XN_STATUS_OK;

    //Read a new frame
    rc = niContext.WaitAnyUpdateAll();

    if(rc != XN_STATUS_OK)
    {
        qDebug() << "Read failed: " << xnGetStatusString(rc);
        return NULL;
    }

    depth.GetMetaData(depthMD);
    image.GetMetaData(imageMD);

    const XnDepthPixel *pDepth = depthMD.Data();

    //Calculate the accumulative histogram( the yellow display )
    xnOSMemSet(pDepthHist, 0, MAX_DEPTH*sizeof(float));

    unsigned int nNumberOfPoints = 0;
    for(XnUInt y=0; y < depthMD.YRes(); ++y)
    {
        for(XnUInt x=0; x < depthMD.XRes(); ++x, ++pDepth)
        {
            if(*pDepth != 0)
            {
                pDepthHist[*pDepth]++;
                nNumberOfPoints++;
            }
        }
    }

    for(int nIndex=1; nIndex < MAX_DEPTH; nIndex++)
    {
        pDepthHist[nIndex] += pDepthHist[nIndex-1];
    }

    if(nNumberOfPoints)
    {
        for(int nIndex=1; nIndex < MAX_DEPTH; nIndex++)
        {
            pDepthHist[nIndex] = (unsigned int)(256 * (1.0f - (pDepthHist[nIndex] / nNumberOfPoints)));
        }
    }

    xnOSMemSet(pTexMap, 0, depthWidth*depthHeight*sizeof(XnRGB24Pixel));

    if(nViewState == DISPLAY_MODE_IMAGE)
    {
        const XnRGB24Pixel *pImageRow = imageMD.RGB24Data();
        XnRGB24Pixel *pTexRow = pTexMap + imageMD.YOffset() * depthWidth;

        for(XnUInt y=0; y < imageMD.YRes(); ++y)
        {
            const XnRGB24Pixel *pImage = pImageRow;
            XnRGB24Pixel *pTex = pTexRow + imageMD.XOffset();

            for(XnUInt x=0; x < imageMD.XRes(); ++x, ++pImage, ++pTex)
            {
                *pTex = *pImage;
            }
            pImageRow += imageMD.XRes();
            pTexRow += depthWidth;
        }
    }



    //Check if we need to draw depth frame to texture
    if(nViewState == DISPLAY_MODE_OVERLAY || nViewState == DISPLAY_MODE_DEPTH)
    {
        const XnDepthPixel *pDepthRow = depthMD.Data();
        XnRGB24Pixel *pTexRow = pTexMap + depthMD.YOffset() * depthWidth;

        const XnRGB24Pixel *pImageRow = imageMD.RGB24Data();
        XnRGB24Pixel *pTexImgRow = pTexMap + imageMD.YOffset() * depthWidth;

        const float alpha = 0.75f;
        for(XnUInt y=0; y < depthMD.YRes(); ++y)
        {
            const XnDepthPixel *pDepth = pDepthRow;
            XnRGB24Pixel *pTex = pTexRow + depthMD.XOffset();

            const XnRGB24Pixel *pImage = pImageRow;
            XnRGB24Pixel *pImgTex = pTexImgRow + imageMD.XOffset();

            for(XnUInt x=0; x < depthMD.XRes(); ++x, ++pDepth, ++pTex, ++pImgTex, ++pImage)
            {

                if(*pDepth > zMin && *pDepth < zMax)
                {
                    int nHistValue = pDepthHist[*pDepth];

                    if(nViewState == DISPLAY_MODE_OVERLAY) {
                        pTex->nRed = alpha*pImage->nRed + (1-alpha)*nHistValue;
                        pTex->nGreen = alpha*pImage->nGreen + (1-alpha)*nHistValue;
                        pTex->nBlue = alpha*pImage->nBlue;
                    } else {
                        pTex->nRed = nHistValue;
                        pTex->nGreen = nHistValue;
                        pTex->nBlue = 0;
                    }
                }
            }

            pImageRow += imageMD.XRes();
            pDepthRow += depthMD.XRes();
            pTexRow += depthWidth;

        }
    }

    frame->pTexMap = pTexMap;
    frame->pRGBMap = imageMD.RGB24Data();
    frame->pDepthMap = depthMD.Data();
    return frame;
}

void NIControl::setup(QString fileName)
{

    if(frame != NULL) {
        delete frame;
        frame = NULL;
    }

    if(pTexMap != NULL) {
        delete [] pTexMap;
        pTexMap = NULL;
    }

    niContext.Shutdown();
    initOpenNI(fileName);
    depthWidth = depthMD.FullXRes();
    depthHeight = depthMD.FullYRes();
    pTexMap = (XnRGB24Pixel*)malloc(depthWidth * depthHeight * sizeof(XnRGB24Pixel));

    frame = new Frame(depthWidth, depthHeight);


}

