#ifndef UTIL_H
#define UTIL_H

#include <vector>
#include <math.h>

#define ALLOCA(TYPE, COUNT) (TYPE *)alloca((COUNT) * sizeof(TYPE))

#ifdef max
#undef max
#endif

#ifdef min
#undef min
#endif

using std::max;

class Filter {
public:
    Filter(float xw, float yw)
        : xWidth(xw), yWidth(yw), invXWidth(1.f/xw), invYWidth(1.f/yw) {
    }

    virtual float evaluate(float x, float y) const = 0;

    const float xWidth, yWidth;
    const float invXWidth, invYWidth;
};

class GaussianFilter : public Filter {
public:
    GaussianFilter(float xw, float yw, float a)
        : Filter(xw, yw), alpha(a), expX( expf(-alpha * xWidth * xWidth)),
          expY( expf(-alpha * yWidth * yWidth)) {
    }

    float evaluate(float x, float y) const {
        return gaussian(x, expX) * gaussian(y, expY);
    }

private:
    float gaussian(float d, float expv) const {
        return std::max(0.f, float(expf(-alpha * d * d) - expv));
    }

private:
    const float alpha;
    const float expX, expY;
};

class MitchellFilter : public Filter {
public:
    MitchellFilter(float b, float c, float xw, float yw)
        : Filter(xw, yw), B(b), C(c) {
    }

    float evaluate(float x, float y) const {
        return mitchell1D(x * invXWidth) * mitchell1D(y * invYWidth);
    }

    float mitchell1D(float x) const {
        x = fabsf(2.f * x);
        if(x > 1.f) {
            return ((-B - 6*C) * x*x*x + (6*B + 30*C) * x*x +
                    (-12*B - 48*C) * x + (8*B + 24*C)) * (1.f/6.f);
        } else {
            return ((12 - 9*B - 6*C) * x*x*x +
                    (-18 + 12*B + 6*C) * x*x +
                    (6 - 2*B)) * (1.f/6.f);
        }
    }

private:
    const float B, C;
};

inline float clamp(float x, float a, float b)
{
    return x < a ? a : (x > b ? b : x);
}


#endif // UTIL_H
